/*
 * Copyright (c) 2014 RWTH Aachen. All rights reserved.
 *
 * http://www.se-rwth.de/
 */
package de.montiarc.extension.examples.processing;

import static com.google.common.base.Preconditions.checkNotNull;
import interfaces2.helper.SymbolTableInterface;

import java.util.ArrayList;
import java.util.List;
import java.util.Stack;

import mc.DSLRoot;
import mc.dsltool.AmbigousException;
import mc.dsltool.CircluarDependencyException;
import mc.umlp.arc._ast.ASTMCCompilationUnit;
import mc.umlp.arcd._ast.ASTArcComponent;
import mc.umlp.arcd._ast.ASTArcElement;
import mc.umlp.arcd._ast.ASTArcSubComponent;
import mc.umlp.arcd.ets.entries.ComponentEntry;
import mc.umlp.arcd.ets.entries.ComponentReferenceEntry;
import mc.umlp.arcd.ets.transform.transformators.IComponentTransformator;
import mc.umlp.arcd.ets.transform.transformators.ISubComponentTransformator;
import mc.umlp.arcd.ets.transform.transformators.TransformatorWithSymtab;

/**
 * Trafo that expands subcomponent to local component definitions in the AST.
 * 
 * @author (last commit) $Author: ahaber $
 * @version $Revision: 3107 $, $Date: 2015-01-23 12:58:56 +0000 (Fri, 23 Jan 2015) $
 * @since 2.5.0
 */
public class ExpandSubcomponentTrafo extends TransformatorWithSymtab
    implements ISubComponentTransformator, IComponentTransformator {
  
  private final Stack<List<ASTArcComponent>> innerComponentsToAdd;
  
  public static final String EXPAND_STEREOTYPE = "expand";
  
  /**
   * Constructor for
   * de.montiarc.extension.examples.processing.ExpandSubcomponentTrafo
   * 
   * @param symtab
   */
  public ExpandSubcomponentTrafo(SymbolTableInterface symtab) {
    super(symtab);
    innerComponentsToAdd = new Stack<List<ASTArcComponent>>();
  }
  
  /**
   * @see mc.umlp.arcd.ets.transform.transformators.ISubComponentTransformator#transform(mc.umlp.arcd._ast.ASTArcSubComponent,
   * mc.umlp.arcd.ets.entries.ComponentReferenceEntry,
   * mc.umlp.arcd.ets.entries.ComponentEntry)
   */
  @Override
  public void transform(ASTArcSubComponent node, ComponentReferenceEntry entry,
      ComponentEntry currentComp) {
    if (node.getStereotype() != null && node.getStereotype().containsStereoValue(EXPAND_STEREOTYPE)) {
      try {
        @SuppressWarnings("unchecked")
        DSLRoot<? extends ASTMCCompilationUnit> root = (DSLRoot<? extends ASTMCCompilationUnit>)
            getSymtab().getModelloader().load(entry.getComponentType().getName());
        checkNotNull(root, "Failed to load component " + entry.getComponentType().getName());
        
        ASTArcComponent copy = root.getAst().getType().deepClone();
        root.getAst().remove_Child(copy);
        innerComponentsToAdd.peek().add(copy);
      }
      catch (CircluarDependencyException e) {
        getDelegator().addErrorToCurrentResource(e.getMessage());
      }
      catch (AmbigousException e) {
        getDelegator().addErrorToCurrentResource(e.getMessage());
      }
    }
  }
  
  /**
   * @see mc.umlp.arcd.ets.transform.transformators.IComponentTransformator#transformAtStart(mc.umlp.arcd._ast.ASTArcComponent,
   * mc.umlp.arcd.ets.entries.ComponentEntry)
   */
  @Override
  public void transformAtStart(ASTArcComponent node, ComponentEntry currentComp) {
    innerComponentsToAdd.push(new ArrayList<ASTArcComponent>());
  }
  
  /**
   * @see mc.umlp.arcd.ets.transform.transformators.IComponentTransformator#transformAtEnd(mc.umlp.arcd._ast.ASTArcComponent,
   * mc.umlp.arcd.ets.entries.ComponentEntry)
   */
  @Override
  public void transformAtEnd(ASTArcComponent node, ComponentEntry currentComp) {
    List<ASTArcComponent> toAdd = innerComponentsToAdd.pop();
    for (ASTArcComponent inner : toAdd) {
      boolean exists = alreadyHasInnerComponent(node, inner);
      if (!exists) {
        node.getBody().getArcElement().add(inner);
      }
    }
    node.clear();
  }
  
  private boolean alreadyHasInnerComponent(ASTArcComponent node, ASTArcComponent inner) {
    boolean result = false;
    for (ASTArcElement elem : node.getBody().getArcElement()) {
      if (elem instanceof ASTArcComponent
          && ((ASTArcComponent) elem).getName().equals(inner.getName())) {
        result = true;
        break;
      }
    }
    return result;
  }
  
}
